from queue import PriorityQueue

class MyPriorityQueue(PriorityQueue):
    def __init__(self):
        PriorityQueue.__init__(self)
        self.counter = 0

    def put(self, item, priority):
        PriorityQueue.put(self, (priority, self.counter, item))
        self.counter += 1

    def get(self, *args, **kwargs):
        _, _, item = PriorityQueue.get(self, *args, **kwargs)
        return item

class Node:
    def __init__(self, board, p=0):
        self.board = board
        self.path = p
        self._estimate_heur()

    def __eq__(self, other):
        return self.board == other.board

    def _final_number_position(self, number):
        if (number == 0):
            return 8
        return number - 1
    
    def _heur_of_number(self, current, final):
        return abs((current % 3) - (final % 3)) + abs((current // 3) - (final // 3))
    
    def _estimate_heur(self):
        self.heur = 0;
        for i in range(1, 9):
            self.heur += self._heur_of_number(self.board.index(i), self._final_number_position(i))

    def _get_row(self, index):
        return index // 3

    def _get_column(self, index):
        return index % 3

    def _swap_by_index(self, board, first_index, second_index):
        board[first_index], board[second_index] = board[second_index], board[first_index]

    def _new_move(self, zero_index, new_zero_index):
        new_board = []
        for i in self.board:
            new_board.append(i)
        self._swap_by_index(new_board, zero_index, new_zero_index)
        return Node(new_board, self.path + 1)

    def print_node(self):
        for i in range(0, 9, 3):
            print("{} {} {}".format(self.board[i], self.board[i + 1], self.board[i + 2]))
        print("Heuristic: {}, Path: {}\n".format(self.heur, self.path))

    def moves(self):
        all_moves = []
        zero_index = self.board.index(0)
        if (self._get_column(zero_index) > 0):
            all_moves.append(self._new_move(zero_index, zero_index - 1))
        if (self._get_column(zero_index) < 2):
            all_moves.append(self._new_move(zero_index, zero_index + 1))
        if (self._get_row(zero_index) > 0):
            all_moves.append(self._new_move(zero_index, zero_index - 3))
        if (self._get_row(zero_index) < 2):
            all_moves.append(self._new_move(zero_index, zero_index + 3))

        return all_moves


def run(start):
    q = MyPriorityQueue()
    visited = []

    q.put(start, start.heur + start.path)
    while(True):
        current = q.get()
        current.print_node()
        if (current.heur == 0):
            print("Finish! Number of moves: {}".format(current.path))
            return True
        visited.append(current)
        for neighbour in current.moves():
            if (neighbour not in visited):
                q.put(neighbour, neighbour.path + neighbour.heur)


board = [
    1, 5, 3,
    4, 2, 6,
    7, 8, 0
    ]

a = Node(board)
run(a)
